#
# Copyright (c) 2004 SASADA Koichi <ko1 at atdot.net>
#
# This program is free software with ABSOLUTELY NO WARRANTY.
# You can re-distribute and/or modify this program under
# the same terms of the Ruby's lisence.
#
#
# $Id: ndk_manager.rb,v 1.31 2004/04/26 01:46:43 ko1 Exp $
# Create : K.S. 04/04/17 17:00:44
#

require 'rice/irc'
require 'ndk_err'
require 'ndk_config'
require 'ndk_state'
require 'ndk_client'

module Nadoka
  Cmd = ::RICE::Command
  Rpl = ::RICE::Reply

  require 'ndk_bot'
  
  class NDK_Manager
    def initialize rc
      @rc = rc
      @clients= []
      @prev_timer = Time.now

      @state   = NDK_State.new self
      reload_config
      @state.config = @config

      host, port, @server_passwd = next_server_info
      
      @server = ::RICE::Connection.new(host, port)
      @connected = false
      @exitting  = false
    end
    attr_reader :state, :connected
    
    def client_count
      @clients.size
    end
    
    def next_server_info
      svinfo = @config.server_list.shift
      @config.server_list.push svinfo
      [svinfo[:host], svinfo[:port], svinfo[:pass]]
    end
    
    def reload_config
      @config = NDK_Config.new(self, @rc)
    end

    def start_server_thread
      @server.regist{|rq, wq|
        Thread.stop
        @rq = rq
        begin
          @connected = false
          server_main_proc
        rescue Exception
          ndk_error $!
          @server_thread.kill  if @server_thread  && @server_thread.alive?
          @clients_thread.kill if @clients_thread && @clients_thread.alive?
        ensure
          @server.close
        end
      }
      @server_thread = Thread.new{
        begin
          @config.slog "Server connection to #{@server.server}:#{@server.port}"
          @server.start(1){|sv|
            sv << Cmd.quit(@config.quit_message) if @config.quit_message
          }
        rescue RICE::Connection::Closed
          host, port, @server_passwd = next_server_info
          @server.server = host
          @server.port   = port
          retry
        rescue Exception
          ndk_error $!
          @clients_thread.kill if @clients_thread && @clients_thread.alive?
        end
      }
    end
    
    def server_main_proc
      # login
      if @server_passwd
        send_to_server Cmd.pass(@server_passwd)
      end
      send_to_server Cmd.nick(@config.nick)
      @state.nick = @config.nick
      
      send_to_server Cmd.user(@config.user,
                              @config.hostname, @config.servername,
                              @config.realname)
      
      # wait welcome message
      while q = recv_from_server
        if q.command == '001'
          break
        elsif q.command == '433'
          # Nickname is already in use.
          nick = @state.nick_succ
          send_to_server Cmd.nick(nick)
        end
      end
      
      # join to default channels
      @config.default_channels.each{|ch|
        send_to_server Cmd.join(ch)
      }

      @connected = true

      if @clients.size == 0
        enter_away
      end
      
      
      # loop
      while q = recv_from_server
        
        case q.command
        when 'PING'
          send_to_server Cmd.pong(q.params[0])
          next
        when 'PRIVMSG'
          if ctcp_message?(q.params[1])
            ctcp_message(q)
          end
        when 'JOIN'
          @state.on_join(sender_of(q), q.params[0])
        when 'PART'
          @state.on_part(sender_of(q), q.params[0])
        when 'NICK'
          @state.on_nick(sender_of(q), q.params[0])
        when 'QUIT'
          @state.on_quit(sender_of(q), q.params[0])
        when 'TOPIC'
          @state.on_topic(sender_of(q), q.params[0], q.params[1])
        when 'MODE'
          @state.on_mode(sender_of(q), q.params[0], q.params[1..-1])
        when '353'
          @state.on_353(q.params[2], q.params[3])
        when '332'
          @state.on_332(q.params[1], q.params[2])
        else
          
        end
        send_to_clients q
        send_to_bot q
        logging q
      end
    end

    def enter_away
      return if @exitting
      
      send_to_server Cmd.away(@config.away_message) if @config.away_message
      
      if @state.nick != @config.away_nick && @config.away_nick
        @state.original_nick = @state.nick
        send_to_server Cmd.nick(@config.away_nick)
      end

      @config.login_channels.each{|ch|
        if @config.channel_info[ch][:part_message]
          send_to_server Cmd.part(ch, @config.channel_info[ch][:part_message])
        else
          send_to_server Cmd.part(ch)
        end
      }
    end

    def leave_away
      return if @exitting

      send_to_server Cmd.away()
      
      if @config.away_nick && @state.original_nick
        send_to_server Cmd.nick(@state.original_nick)
        @state.original_nick = nil
      end

      @config.login_channels.each{|ch|
        send_to_server Cmd.join(ch)
      }
    end
    
    
    def start_clients_thread
      @clients_thread = Thread.new{

        begin
          @cserver = TCPServer.new(@config.client_server_host,
                                   @config.client_server_port)
          @config.slog "Open Client Server Port: #{@cserver.addr.join(' ')}"
          
          while true
            # wait for client connections
            Thread.start(@cserver.accept){|cc|
              client = nil
              begin
                if !@config.acl_object || @config.acl_object.allow_socket?(cc)
                  client = NDK_Client.new(@config, cc, self)
                  @clients << client
                  client.start
                else
                  @config.slog "ACL denied: #{cc.peeraddr.join(' ')}"
                end
              ensure
                @clients.delete client
                invoke_event :enter_away, client_count
                cc.close unless cc.closed?
              end
            }
          end
        rescue
          ndk_error $!
        ensure
          @clients.each{|cl|
            cl.kill
          }
          if @cserver
            @config.slog "Close Client Server Port: #{@cserver.addr.join(' ')}"
            @cserver.close unless @cserver.closed?
          end
          @server_thread.kill if @server_thread.alive?
        end
      }
    end

    def about_me?(msg)
      /^#{@state.nick}!/ =~ msg.prefix
    end
    
    def start
      start_server_thread
      start_clients_thread

      begin
        @server_thread.join
      rescue Interrupt
        @exitting = true
      ensure
        @server_thread.kill  if @server_thread  && @server_thread.alive?
        @clients_thread.kill if @clients_thread && @clients_thread.alive?
        
        @server.close if @server
      end
    end
    
    def send_to_server msg
      @config.dlog "[>S] #{msg}"
      @server << msg
    end
    
    def recv_from_server
      while q = @rq.pop
        # Timer event?
        t = Time.now
        if (t.to_i - @prev_timer.to_i) > 60
          @prev_timer = t
          send_to_bot :timer, t
        end

        if q.kind_of? Array
          case q[0]
          when :reload_config
            # q[1] must be client object
            reload_config
            q[1] << Cmd.notice(@state.nick, "configuration is reloaded")
            
          when :quit_program
            @exitting = true
            Thread.main.raise NDK_QuitProgram
            
          when :restart_program
            @exitting = true
            Thread.main.raise NDK_RestartProgram
            
          when :invoke_bot
            # q[1], q[2] are message and argument
            send_to_bot q[1], q[2]

          when :enter_away
            if q[1] == 0
              enter_away
            end
            
          when :leave_away
            if q[1] == 1
              leave_away
            end
            
          end
          next
        end
        
        ##
        case q.command
        when 'PING'
          @server << Cmd.pong(q.params[0])
        when 'PONG'
          # ignore
        else
          @config.dlog "[<S] #{q}"
          return q
        end
      end
    end

    def invoke_event ev, *arg
      arg.unshift ev
      @rq && (@rq << arg)
    end
    
    # server -> clients
    def send_to_clients msg
      @clients.each{|cl|
        cl << msg
      }
    end

    # clientA -> other clients
    # bot     -> clients
    def send_to_clients_otherwise msg, elt
      @clients.each{|cl|
        if cl != elt
          cl.add_prefix(msg) unless msg.prefix
          cl << msg
        end
      }
      invoke_event :invoke_bot, msg if elt
      logging msg
    end
    
    def ctcp_message? arg
      arg[0] == 1
    end
    
    def ctcp_message msg
      if /\001(.+)\001/ =~ msg.params[1]
        ctcp_cmd = $1
        case ctcp_cmd
        when 'VERSION'
          send_to_server Cmd.notice(sender_of(msg), "\001VERSION #{Nadoka.version}\001")
        when 'TIME'
          send_to_server Cmd.notice(sender_of(msg), "\001TIME #{Time.now}\001")
        else
          
        end
      end
    end
    
    def sender_of msg
      if /^([^!]+)\!?/ =~ msg.prefix.to_s
        $1
      else
        @state.nick
      end
    end

    class PrefixObject
      def initialize prefix
        parse_prefix prefix
        @prefix = prefix
      end
      attr_reader :nick, :user, :host, :prefix

      def parse_prefix prefix
        if /^(.+?)\!(.+?)@(.+)/ =~ prefix.to_s
          # command
          @nick, @user, @host = $1, $2, $3
        else
          # server reply
          @nick, @user, @host = nil, nil, prefix
        end
      end

      def to_s
        @prefix
      end
    end
    
    def make_prefix_object msg
      prefix = msg.prefix
      if prefix
        PrefixObject.new(prefix)
      else
        if /^d+$/ =~ msg.command
          PrefixObject.new("nadokatan")
        else
          PrefixObject.new("#{@state.nick}!#{@config.user}@nadokatan")
        end
      end
    end
    
    # dispatch to bots
    def send_to_bot msg, arg = nil

      selector = 'on_' +
      if msg.respond_to? :command
        if /^\d+$/ =~ msg.command
          # reply
          prefix = make_prefix_object msg
          RICE::Reply::Replies_num_to_name[msg.command]
        else
          # command
          prefix = make_prefix_object msg
          msg.command.downcase
        end
      else
        prefix = nil
        msg.to_s
      end
      
      
      @config.bots.each{|bot|
        begin
          if bot.respond_to? selector
            unless prefix
              bot.__send__(selector, arg)
            else
              bot.__send__(selector, prefix, *msg.params)
            end
          end

          if prefix && bot.respond_to?(:on_every_message)
            bot.__send__(:on_every_message, prefix, msg.command, *msg.params)
          end
          
        rescue NDK_BotBreak
          break
          
        rescue Exception
          ndk_error $!
        end
      }
    end

    # logging
    def logging msg
      user = sender_of(msg)
      ch   = msg.params[0]
      case msg.command
      when 'PRIVMSG'
        str = "<#{ch}:#{user}> #{msg.params[1]}"
        @config.clog ch, str
        
      when 'NOTICE'
        str = "{#{ch}:#{user}} #{msg.params[1]}"
        @config.clog ch, str
        
      when 'JOIN', 'PART', 'NICK', 'QUIT'
        # ignore
        
      when /^\d+/
        str = msg.command + ' ' + msg.params.join(' ')
        @config.log str
        
      else
        # other command
        @config.log msg.to_s
      end
    end

    def ndk_error err
      $stderr.puts err
      $stderr.puts err.backtrace.join(' ')
      @config.slog err
      @config.slog err.backtrace.join(' ')
    end
    
  end
end


